"""

Low-level methods for managing applications in MiCADO

"""
import json

from micado.types import ApplicationInfo
from micado.exceptions import detailed_raise_for_status

class ApplicationMixin:
    def applications(self):
        """Lists the currently running applications

        Returns:
            list: Current application IDs
        """
        url = self._url("/applications/")
        resp = self.get(url)
        detailed_raise_for_status(resp)
        return resp.json()["applications"]

    def inspect_app(self, app_id):
        """Fetches detailed info on an application

        Args:
            app_id (string): The ID of the application

        Returns:
            dict: Info on the given application
        """
        url = self._url(f"/applications/{app_id}/")
        resp = self.get(url)
        detailed_raise_for_status(resp)
        return resp.json()

    def create_app(
        self, app_id=None, adt=None, url=None, params=None, dryrun=False, file=None
    ):
        """Creates/deploys an application in MiCADO

        Args:
            app_id (string, optional): The ID of the application to deploy.
                Defaults to None.
            adt (dict, optional): YAML dict of the application description
                template. Required if no URL or file provided. Defaults to None.
            url (string, optional): URL to YAML document of the ADT.
                Required if no adt(dict) or file is provided. Defaults to None.
            params (dict, optional): Dict of TOSCA input values.
                Defaults to None.
            dryrun (bool, optional): Flag to skip execution of components.
                Defaults to False.
            file (string, optional): YAML or CSAR file to submit. Required
                if no adt(dict) or URL is provided. Defaults to None.

        Raises:
            TypeError: If no ADT/URL data is passed in

        Returns:
            dict: ID and status of deployed application
        """
        if not adt and not url and not file:
            raise TypeError("Either adt or url or file is required.")
        if app_id:
            endpoint = self._url(f"/applications/{app_id}/")
        else:
            endpoint = self._url("/applications/")

        json_data = ApplicationInfo(adt, url, params, dryrun)
        if file:
            form_data = {
                k: json.dumps(v) if isinstance(v, dict) 
                else str(v) 
                for k, v in json_data.items()
            }
            file_data = {"adt": file}
            resp = self.post(endpoint, files=file_data, data=form_data)
        else:
            resp = self.post(endpoint, json=json_data)
        detailed_raise_for_status(resp)
        return resp.json()

    def delete_app(self, app_id, force=False):
        """Delete an application in MiCADO

        Args:
            app_id (string): ID of application to delete
            force (bool, optional): Ignore errors. Defaults to False.

        Returns:
            dict: ID and status of deletion
        """
        url = self._url(f"/applications/{app_id}/")
        json_data = {"force": force}
        resp = self.delete(url, json=json_data)
        if not force:
            detailed_raise_for_status(resp)
        return resp.json()

    def _destroy(self):
        """Deletes all application in MiCADO

        This should normally only be called by a launcher that
        is ready to destroy the entire MiCADO stack
        """
        app_ids = self.applications()
        for app in app_ids:
            self.delete_app(app, force=True)



import logging
import logging.config
import os
import paramiko
import requests
import socket
import subprocess
import time
import urllib3
from pathlib import Path
from requests.adapters import Retry, HTTPAdapter

from micado.installer.ansible.playbook import Playbook
from micado.exceptions import MicadoException
from micado.utils.utils import DataHandling, generate_password
from ruamel.yaml import YAML

DEFAULT_PATH = Path.home() / ".micado-cli"
DEFAULT_VERS = "v0.11.0"
API_VERS = "v2.0"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
micado_cli_dir = Path(os.environ.get("MICADO_DIR", DEFAULT_PATH))
micado_cli_dir.mkdir(parents=True, exist_ok=True)
ch = logging.StreamHandler()
fh = logging.handlers.RotatingFileHandler(
    filename=str(micado_cli_dir / "micado-cli.log"),
    mode="a",
    maxBytes=52428800,
    backupCount=3,
)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s : %(message)s")
ch.setFormatter(formatter)
fh.setFormatter(formatter)
logger.addHandler(ch)
logger.addHandler(fh)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class AnsibleInstaller:
    micado_version = os.environ.get("MICADO_VERS", DEFAULT_VERS)
    api_version = os.environ.get("API_VERS", API_VERS)
    home = str(Path(os.environ.get("MICADO_DIR", DEFAULT_PATH))) + "/"

    def deploy(
        self,
        micado,
        micado_user="admin",
        micado_password=None,
        terraform=True,
        occopus=False,
        **kwargs,
    ):
        instance_ip = micado.ip
        micado_id = micado.id

        logger.info("Check instance availability...")
        self._check_availability(instance_ip)

        logger.info("Generating playbook inputs...")
        micado_password = micado_password or generate_password()
        hosts = self._generate_inventory(instance_ip)
        extravars = self._generate_extravars(
            micado_user, micado_password, terraform, occopus
        )

        logger.info("Running playbook...")
        self._run_playbook(micado_id, hosts, extravars)
        self._check_submitter(instance_ip, micado_user, micado_password)
        logger.info("MiCADO deployed!")

        self._get_self_signed_cert(instance_ip, micado_id)
        self._store_data(micado_id, self.api_version, micado_user, micado_password)
        logger.info(f"MiCADO ID is: {micado_id}")
        
    def _check_submitter(self, instance_ip, user, passw):
        """Check the submitter endpoint is returning 200"""
        self._check_port_availability(instance_ip, 443)
        s = requests.Session()
        s.auth = (user, passw)
        s.verify = False
        
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[500, 502, 503, 504])
        s.mount("https://", HTTPAdapter(max_retries=retries))
        s.get(f"https://{instance_ip}/toscasubmitter/v2.0/applications/")

    def _check_availability(self, instance_ip):
        """Perform availability checks"""
        self._check_port_availability(instance_ip, 22)
        self._remove_know_host()
        self._get_ssh_fingerprint(instance_ip)
        self._check_ssh_availability(instance_ip)

    def _run_playbook(self, micado_id, hosts, extravars):
        """Run the playbook"""
        playbook = Playbook(self.micado_version, micado_id, self.home)
        runner = playbook.run(hosts, extravars)
        if runner.rc == 0:
            logger.info("Playbook complete.")
        else:
            msg = "\n".join([event["stdout"] for event in list(runner.events)[-5:]])
            logger.error(msg)
            raise MicadoException(msg)

    def _generate_inventory(self, ip):
        """Generate hosts info for Playbook

        Args:
            ip (string): MiCADO IP
        """
        host_dict = {}
        host_dict[
            "ansible_ssh_private_key_file"
        ] = f"{self.home}micado_cli_config_priv_key"
        host_dict["ansible_host"] = ip
        host_dict["ansible_user"] = "ubuntu"
        hosts = {"all": {"hosts": {"micado": host_dict}}}

        return hosts

    def _generate_extravars(self, micado_user, micado_password, terraform, occopus):
        """Configure ansible-micado, with credentials, etc...

        Args:
            micado_user (string): User defined MiCADO user
            micado_password (string): User defined MiCADO password
            terraform (boolean): Terraform enabled
            occopus (boolean): Occopus enabled
        """

        security_dict = self._generate_credential_data(micado_user, micado_password)

        extra_variables = {
            "cloud_cred_path": str(micado_cli_dir / "credentials-cloud-api.yml"),
            "registry_cred_path": str(micado_cli_dir / "credentials-registries.yml"),
            "enable_terraform": terraform,
            "enable_occopus": occopus,
            "security": security_dict,
        }

        return extra_variables

    def _generate_credential_data(self, micado_user, micado_password):
        """Create MiCADO credential file.

        Args:
            micado_user (string): User defined MiCADO user
            micado_password ([type]): User defined MiCADO password
        """
        logger.info("Loading MiCADO credentials...")

        auth_dict = {
            "authentication": {"username": micado_user, "password": micado_password}
        }

        micado_cred_path = micado_cli_dir / "credentials-micado.yml"
        if not micado_cred_path.is_file():
            return auth_dict

        with open(micado_cred_path, "r") as f:
            yaml = YAML()
            credential_dict = yaml.load(f)
            credential_dict.update(auth_dict)

        return credential_dict

    def _check_port_availability(self, ip, port):
        """Check the given port availability.

        Args:
            ip (string): IP address of the VM
            port (string): Port number

        Raises:
            Exception: When timeout reached
        """
        logger.info("Check {} port availability...".format(port))
        attempts = 0
        sleep_time = 2
        max_attempts = 1000
        result = None
        logger.debug(
            "IP: {} \tPort: {} \tsleeptime:{}".format(
                ip,
                port,
                sleep_time,
            )
        )
        while attempts < max_attempts and result != 0:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            socket.setdefaulttimeout(1)
            result = s.connect_ex((ip, port))
            s.close()
            attempts += 1
            if result == 0:
                logger.info("{} port is available...".format(port))
                break
            logger.debug(
                "attempts:{}/{} Still no answer. Try again {} seconds later...".format(
                    attempts, max_attempts, sleep_time
                )
            )
            time.sleep(sleep_time)

        if attempts == max_attempts:
            raise Exception(
                "{} second passed, and still cannot reach {}.".format(
                    attempts * sleep_time, port
                )
            )

    def _remove_know_host(self):
        """Remove known_host file"""
        known_hosts = str(Path.home()) + "/.ssh/known_hosts"
        if not os.path.isfile(known_hosts):
            return
        with open(known_hosts) as file:
            all_lines = file.readlines()
        with open(known_hosts + ".old", "a") as file2:
            file2.writelines(all_lines)
        os.remove(known_hosts)

    def _get_ssh_fingerprint(self, ip):
        """Get SSH fingerprint

        Args:
            ip (string): Target IP address
        """
        known_hosts = str(Path.home()) + "/.ssh/known_hosts"
        result = subprocess.run(
            ["ssh-keyscan", "-H", ip],
            shell=False,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        with open(known_hosts, "a") as f:
            f.writelines(result.stdout.decode())

    def _get_self_signed_cert(self, ip, id):
        """Get MiCADO self signed SSL

        Args:
            ip (string): Target IP
            id (string): UUID of the VM
        """
        logger.info("Get MiCADO self_signed cert")

        key = paramiko.RSAKey.from_private_key_file(
            f"{self.home}micado_cli_config_priv_key"
        )
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=ip, username="ubuntu", pkey=key)
        sftp = ssh.open_sftp()
        sftp.get("/var/lib/micado/zorp/config/ssl.pem", f"{self.home}{id}-ssl.pem")
        sftp.close()
        ssh.close()

    def _store_data(self, server_id, api_version, micado_user, micado_password):
        """Persist configuration specific data

        Args:
            server_id (string): UUID of the server
            api_version (string): Toscasubmitter API version
            micado_user (string): MiCADO username
            micado_password (string): MiCADO password
        """
        cert_path = f"{self.home}{server_id}-ssl.pem"
        DataHandling.update_data(
            self.home + "data.yml",
            server_id,
            api_version=api_version,
            micado_user=micado_user,
            micado_password=micado_password,
            cert_path=cert_path,
        )

    def get_api_version(self):
        """
        Return the MiCADO Submitter API version. Only v2.0 supported.

        Returns:
            string: MiCADO Submitter API version
        """

        return self.api_version

    def _check_ssh_availability(self, ip):
        """Check SSH availability

        Args:
            ip (string): Target IP

        Raises:
            Exception: When timeout reached
        """
        attempts = 0
        sleep_time = 2
        max_attempts = 100
        while attempts < max_attempts:
            result = subprocess.run(
                [
                    "ssh",
                    "-i",
                    self.home + "micado_cli_config_priv_key",
                    "ubuntu@" + ip,
                    "ls -lah",
                ],
                shell=False,
                stdin=None,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=False,
            )
            if result.returncode == 0:
                logger.debug("SSH connection available...")
                break
            logger.debug(result.stderr.decode())
            attempts += 1
            logger.debug(
                "attempts:{}/{} Cloud-init still running. Try again {} second later".format(
                    attempts + 1, max_attempts, sleep_time
                )
            )
            time.sleep(sleep_time)

        if attempts == max_attempts:
            raise Exception(
                "{} second passed, and still cannot reach SSH.".format(
                    attempts * sleep_time
                )
            )



#!/usr/bin/env python

import logging
import logging.config
import os
import uuid
from pathlib import Path
from urllib.parse import urlparse
from time import sleep
import itertools as it
import xml.dom.minidom
from xml.dom.minidom import parseString
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement, tostring
import xml.etree.ElementTree as ET
from dicttoxml import dicttoxml
from collections import OrderedDict
import requests, json
import base64
from micado.types.micado import MicadoInfo
from micado.exceptions import MicadoException
from micado.utils.utils import DataHandling, SSHKeyHandling
import ruamel.yaml as yaml

"""
Low-level methods for handling a MiCADO node with CloudBroker API
"""
DEFAULT_PATH = Path.home() / ".micado-cli"
DEFAULT_VERS = "0.9.1-rev1"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
micado_cli_dir = Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))
micado_cli_dir.mkdir(parents=True, exist_ok=True)
ch = logging.StreamHandler()
fh = logging.handlers.RotatingFileHandler(
    filename=str(micado_cli_dir / "micado-cli.log"),
    mode="a",
    maxBytes=52428800,
    backupCount=3,
)
ch.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s : %(message)s')
ch.setFormatter(formatter)
fh.setFormatter(formatter)
logger.addHandler(ch)
logger.addHandler(fh)


class CloudBrokerLauncher:
    """
    For launching a MiCADO node with CloudBroker API
    """
    home = str(Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))) + '/'

    def launch(self, auth_url, deployment_id=None, instance_type_id=None, key_pair_id=None, firewall_rule_set_id=None, **kwargs):
        """
        Create the MiCADO node

        Args:
            auth_url ([type]): [description]
            deployment_id ([type]): [description]
            instance_type_id ([type]): [description]
            key_pair_id ([type]): [description]
            firewall_rule_set_id ([type]): [description]

        Returns:
            MicadoInfo: Dataclass with MiCADO ID and IP
        """
        instanceID = ""

        try:
            pub_key = SSHKeyHandling.get_pub_key(self.home)
            if deployment_id is None:
                raise MicadoException("Can't find deployment id!")
            if instance_type_id is None:
                raise MicadoException("Can't find instance type id!")
            if key_pair_id is None:
                raise MicadoException("Can't find key pair id!")
            if firewall_rule_set_id is None:
                raise MicadoException("Can't find firewall rule set id!")

            descr = {}
            descr['deployment_id'] = deployment_id
            descr['instance_type_id'] = instance_type_id
            descr['key_pair_id'] = key_pair_id
            descr['firewall_rule_set_id'] = firewall_rule_set_id
            descr.setdefault('disable_autostop', 'true')
            descr.setdefault('isolated', 'true')
            name_id = uuid.uuid1()
            name = 'MiCADO-{}'.format(name_id.hex)
            descr['name'] = name

            logger.info('Creating CloudBroker VM...')
            cloud_init_config = """
            #cloud-config

            ssh_authorized_keys:
            - {}
            """.format(pub_key)

            descr['cloud-init'] = base64.b64encode(cloud_init_config.encode('utf-8')).decode('utf-8')
            descr['cloud-init-b64'] = 'true'

            logger.debug("XML to pass to CloudBroker: %s", dicttoxml(descr, custom_root='instance', attr_type=False))
            r = requests.post(auth_url + '/instances.xml',
                              dicttoxml(descr, custom_root='instance', attr_type=False),
                              auth=self.get_auth(),
                              headers={'Content-Type': 'application/xml'})            

            logger.debug('CloudBroker instance create response status code %d, response: %s', r.status_code, r.text)
            if (r.status_code == 201):
                DOMTree = xml.dom.minidom.parseString(r.text)
                instance = DOMTree.documentElement
                instanceID = instance.getElementsByTagName('id')[0].childNodes[0].data
                logger.info("CloudBroker instance started, instance id: %s", instanceID)
                mstate = "starting"
                while mstate != "running":
                    sleep(5)
                    instance = self.get_instance(auth_url, instanceID)
                    mstate = self.getTagText(instance.getElementsByTagName('status').item(0).childNodes)
                floating_ip_address = self.getTagText(instance.getElementsByTagName('external-ip-address').item(0).childNodes)
                self._persist_data(floating_ip_address, instanceID, auth_url, deployment_id, instance_type_id, key_pair_id, firewall_rule_set_id)
                return MicadoInfo(instanceID, floating_ip_address)
            else:
                errormsg = 'Failed to create CloudBroker instance, request status code {0}, response: {1}'.format(r.status_code, r.text)
                logger.debug(errormsg)
                raise Exception(errormsg)

        except MicadoException as e:
            logger.error(f"Exception cought: {e}")
            raise
        except Exception as e:
            logger.error(f"Exception cought: {e}")
            raise

    def delete(self, id):
        """
        Destroy the existing MiCADO VM.
        Args:
            id (string): The MiCADO UUID.
        Raises:
            MicadoException: Missing or incorrect data.
        """
        try:
            auth_url = None
            content = None
            with open(self.home + 'data.yml', mode='r') as f:
                content = yaml.safe_load(f)
            search = [i for i in content["micados"] if i.get(id, None)]
            if not search:
                logger.debug(
                    "This {} ID can not find in the data file.".format(id))
                pass
            else:
                logger.debug("Remove {} record".format(search))
                auth_url = search[0][id]["auth_url"]
                content["micados"].remove(search[0])
                with open(self.home + 'data.yml', mode='w') as f:
                    yaml.dump(content, f)

            r = requests.put(auth_url + '/instances/' + id + '/stop.xml', auth=self.get_auth())
            logger.info('Dropping node {}'.format(id))
            if os.path.isfile(self.home + id + '-ssl.pem'):
                logger.info("remove {}-ssl.pem".format(self.home + id))
                os.remove(self.home + id + '-ssl.pem')
            return "Destroyed"
        except MicadoException as e:
            logger.error(f"Exception cought: {e}")

    def _get_credentials(self):
        """
        Read CloudBroker credentials from file.
        Usage : auth_data = self._get_credentials()
        """
        with open(self.home + "credentials-cloud-api.yml", "r") as stream:
            temp = yaml.safe_load(stream)

        resources = temp.get("resource", {})
        for resource in resources:
            if resource.get("type") == "cloudbroker":
                return resource.get("auth_data")

    def get_auth(self):
        auth_data = self._get_credentials()
        if (not auth_data) or (not "email" in auth_data) or (not "password" in auth_data):
            errormsg = "Cannot find credentials for CloudBroker. Please specify"
            logger.debug(errormsg)
        return (auth_data['email'], auth_data['password'])

    def _persist_data(self, floating_ip_address, instanceID, auth_url, deployment_id, instance_type_id, key_pair_id, firewall_rule_set_id):
        """
        """
        endpoint = f'https://{floating_ip_address}/toscasubmitter'
        file_location = self.home + "data.yml"
        DataHandling.persist_data(path=file_location,
                                  server_id=instanceID,
                                  ip=floating_ip_address,
                                  auth_url=auth_url,
                                  deployment_id=deployment_id,
                                  instance_type_id=instance_type_id,
                                  key_pair_id=key_pair_id,
                                  firewall_rule_set_id=firewall_rule_set_id,
                                  endpoint=endpoint)

    def get_instance(self, auth_url, instanceid):
        attempt = 0
        stime = 1
        while attempt < 5:
            query_str = auth_url + '/instances/' + instanceid + '.xml'
            r = requests.get(query_str, auth=self.get_auth())
            if (r.status_code != 200):
                logger.debug('CloudBroker API call failed! query: %s, status code %d, response: %s', query_str, r.status_code, r.text)
            else:
                DOMTree = xml.dom.minidom.parseString(r.text)
                instance = DOMTree.documentElement
                if 0 != instance.getElementsByTagName('id').length:
                    return instance
                else:
                    logger.debug('CloudBroker API returned incorrect answer! No instance id is found.query: %s, status code %d, response: %s', query_str, r.status_code, r.text)
            sleep(stime)
            stime = stime * 2
            attempt += 1
            logger.debug('Retry calling the CloudBroker API...')
        errormsg = 'Error in querying instance \'{0}\' {1} times through CloudBroker API at \'{2}\'.'.format(
               str(instanceid), str(attempt), auth_url)
        logger.debug(errormsg)
        raise Exception(errormsg)

    def getTagText(self, nodelist):
        rc = []
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc.append(node.data)
        return ''.join(rc)



#!/usr/bin/env python

import logging
import logging.config
import os
import random
import subprocess
import time
import uuid
from pathlib import Path

import requests
from keystoneauth1 import session
from micado.exceptions import MicadoException
from micado.utils.utils import DataHandling, SSHKeyHandling
from novaclient import client as nova_client
from ruamel.yaml import YAML

import openstack
from openstack import connection

from .auth import AUTH_TYPES
from micado.types.micado import MicadoInfo

"""Low-level methods for handling a MiCADO node with OpenStackSDK

"""
DEFAULT_PATH = Path.home() / ".micado-cli"
DEFAULT_VERS = "0.9.1-rev1"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
micado_cli_dir = Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))
micado_cli_dir.mkdir(parents=True, exist_ok=True)
ch = logging.StreamHandler()
fh = logging.handlers.RotatingFileHandler(
    filename=str(micado_cli_dir / "micado-cli.log"),
    mode="a",
    maxBytes=52428800,
    backupCount=3,
)
ch.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s : %(message)s')
ch.setFormatter(formatter)
fh.setFormatter(formatter)
logger.addHandler(ch)
logger.addHandler(fh)


class OpenStackLauncher:
    """For launching a MiCADO node with OpenStackSDK

    """
    home = str(Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))) + '/'

    def launch(self, auth_url, image, flavor, network, keypair, security_group='all', region=None,
               user_domain_name='Default', project_id=None, **kwargs):
        """Create the MiCADO node

        Args:
            auth_url ([type]): [description]
            image ([type]): [description]
            flavor ([type]): [description]
            network ([type]): [description]
            keypair ([type]): [description]
            security_group (str, optional): [description]. Defaults to 'all'.
            region ([type], optional): [description]. Defaults to None.
            user_domain_name (str, optional): [description]. Defaults to 'Default'.
            project_id ([type], optional): [description]. Defaults to None.

        Raises:
            MicadoException: [description]
            MicadoException: [description]
            MicadoException: [description]
            MicadoException: [description]
            MicadoException: [description]
            MicadoException: [description]

        Returns:
            MicadoInfo: Dataclass with MiCADO ID and IP
        """
        try:
            pub_key = SSHKeyHandling.get_pub_key(self.home)
            conn, conn_nova = self._get_connection(
                auth_url, region, project_id, user_domain_name)
            image = conn.get_image(image)
            flavor = conn.get_flavor(flavor)
            network = conn.get_network(network)
            keypair = conn.get_keypair(keypair)
            security_group = conn.get_security_group(security_group)
            if image is None:
                raise MicadoException("Can't find image!")
            if flavor is None:
                raise MicadoException("Can't find flavor!")
            if network is None:
                raise MicadoException("Can't find network!")
            if keypair is None:
                raise MicadoException("Can't find keypair!")
            if security_group is None:
                raise MicadoException("Can't find security_group!")
            unused_ip = self.get_unused_floating_ip(conn)
            if len(unused_ip) < 1:
                raise MicadoException("Can't find availabe floating IP!")
            ip = random.choice(unused_ip)
            logger.info('Creating VM...')
            cloud_init_config = """
            #cloud-config

            ssh_authorized_keys:
            - {}
            """.format(pub_key)
            name_id = uuid.uuid1()
            server = conn_nova.servers.create(
                'MiCADO-{}'.format(name_id.hex),
                image.id,
                flavor.id,
                security_groups=[security_group.id],
                nics=[{"net-id": network.id}],
                key_name=keypair.name,
                userdata=cloud_init_config)
            # server = conn.compute.create_server(
            #     name='MiCADO-{}'.format(name_id.hex), image_id=image.id, flavor_id=flavor.id,
            # key_name=keypair.name, userdata=cloud_init_config, timeout=300,
            # networks=[{"uuid": network.id}], security_groups=[{"name":
            # security_group.id}])
            logger.info('The VM {} starting...'.format(server.id))
            server = conn.get_server(server.id)
            logger.info('Waiting for running state, and attach {} floating ip...'.format(
                ip.floating_ip_address))
            conn.wait_for_server(server, auto_ip=False,
                                 ips=ip.floating_ip_address, timeout=600)
            self._persist_data(ip.floating_ip_address, server.id,
                               auth_url, region, project_id, user_domain_name)
            return MicadoInfo(server.id, ip.floating_ip_address)
        except MicadoException as e:
            logger.error(f"Exception cought: {e}")
            raise
        except Exception as e:
            logger.error(f"Exception cought: {e}")
            if 'server' in locals():
                conn.delete_server(server.id)
                logger.info(f"{server.id} VM dropped.")
            raise

    def delete(self, id):
        """Destroy the existing MiCADO VM.

        Args:
            id (string): The MiCADO UUID.

        Raises:
            MicadoException: Missing or incorrect data.
        """
        try:
            auth_url = None
            region_name = None
            project_id = None
            user_domain_name = None
            yaml = YAML()
            content = None
            with open(self.home + 'data.yml', mode='r') as f:
                content = yaml.load(f)
            search = [i for i in content["micados"] if i.get(id, None)]
            if not search:
                logger.debug(
                    "This {} ID can not find in the data file.".format(id))
                pass
            else:
                logger.debug("Remove {} record".format(search))
                auth_url = search[0][id]["auth_url"]
                region_name = search[0][id]["region_name"]
                project_id = search[0][id]["project_id"]
                user_domain_name = search[0][id]["user_domain_name"]
                content["micados"].remove(search[0])
                with open(self.home + 'data.yml', mode='w') as f:
                    yaml.dump(content, f)
            conn, _ = self._get_connection(
                auth_url, region_name, project_id, user_domain_name)
            if conn.get_server(id) is None:
                raise MicadoException("{} is not a valid VM ID!".format(id))
            conn.delete_server(id)
            logger.info('Dropping node {}'.format(id))
            if os.path.isfile(self.home + id + '-ssl.pem'):
                logger.info("remove {}-ssl.pem".format(self.home + id))
                os.remove(self.home + id + '-ssl.pem')
            return "Destroyed"
        except MicadoException as e:
            logger.error(f"Exception cought: {e}")

    def _get_credentials(self):
        """Read credential from file.

        Raises:
            TypeError: Missing or incorrect credential data.

        Returns:
            Authenticator: specific authenticator object
        """
        with open(self.home + "credentials-cloud-api.yml", "r") as stream:
            yaml = YAML()
            auth_data = yaml.load(stream)

        resources = {}
        for resource in auth_data["resource"]:
            resources[resource["type"]] = resource["auth_data"]

        nova = {
            key: val
            for key, val
            in resources.get("nova", {}).items()
            if val
        }

        errors = []
        for auth in AUTH_TYPES:
            try:
                return auth(**nova)
            except TypeError as error:
                errors.append(f"{auth.__name__}.{error}")

        errors = "\n" + "\n".join(errors)
        raise TypeError(f"Incomplete/ambiguous credentials: {errors}")

    def get_unused_floating_ip(self, conn):
        """Return unused ip.

        Args:
            conn ([type]): OpenStack connection

        Returns:
            dict: Unused IP
        """
        return [addr for addr in conn.list_floating_ips()
                if addr.attached == False]

    def _get_connection(
        self, auth_url, region_name, project_id, user_domain_name
    ):
        """Create OpenStack connection.

        Args:
            auth_url (string): Authentication URL for the NOVA
                resource.
            region_name (string, optional): Name of the region resource.
                Defaults to None.
            project_id (string, optional): ID of the project resource.
                Defaults to None.
            user_domain_name (string, optional): Define the user_domain_name.
                Defaults to 'Default'

        Raises:
            Exception: Project ID missing

        Returns:
            tuple: OpenStackSDK connection, and nova_client Connection
        """
        logger.info("Pulling credentials...")
        authenticator = self._get_credentials()
        authenticator.project_id = project_id
        authenticator.user_domain_name = user_domain_name
        authenticator.auth_url = auth_url

        logger.info("Authenticating with OpenStack...")
        auth = authenticator.authenticate()
        sess = session.Session(auth=auth)
        return (
            connection.Connection(
                region_name=region_name,
                session=sess,
                compute_api_version="2",
                identity_interface="public",
            ),
            nova_client.Client(2, session=sess, region_name=region_name),
        )

    def _persist_data(self, ip, server_id, auth_url,
                      region_name, project_id, user_domain_name):
        """
        """
        endpoint = f'https://{ip}/toscasubmitter'
        file_location = self.home + "data.yml"
        DataHandling.persist_data(path=file_location,
                                  server_id=server_id,
                                  ip=ip,
                                  auth_url=auth_url,
                                  region_name=region_name,
                                  project_id=project_id,
                                  user_domain_name=user_domain_name,
                                  endpoint=endpoint)



import ast


class ApplicationInfo(dict):
    def __init__(self, adt=None, url=None, params=None, dryrun=False):
        """Creates required JSON parameters for create_app

        Args:
            adt (dict, optional): YAML dict of Application Description
                Template. Required is URL is empty. Defaults to None.
            url (string, optional): URL of YAML ADT. Required if ADT is empty.
                Defaults to None.
            params (dict, optional): TOSCA input parameters. Defaults to {}.
            dryrun (bool, optional): Flag to skip execution of components.
                Defaults to False.
        """
        if adt:
            self["adt"] = adt
        elif url:
            self["url"] = url
        if isinstance(params, str):
            self["params"] = ast.literal_eval(params)
        elif isinstance(params, dict):
            self["params"] = params
        self["dryrun"] = dryrun



from setuptools import setup, find_packages
from pathlib import Path

with open("README.md") as file:
    long_description = file.read()

REQUIREMENTS = [
    "requests",
    "ruamel.yaml",
    "paramiko",
    "pycryptodome",
    "python-novaclient",
    "openstacksdk",
    "ansible",
    "ansible-runner",
    "click",
    "dicttoxml",
]

setup(
    name="micado-client",
    version="0.12.3",
    description="A Python Client Library for MiCADO",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Márk Emődi & Jay DesLauriers",
    python_requires=">=3.8",
    url="https://github.com/micado-scale/micado-client",
    packages=find_packages(exclude=["tests"]),
    project_urls={
        "Bug Tracker": "https://github.com/micado-scale/micado-client/issues",
    },
    install_requires=REQUIREMENTS,
    license="Apache 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={
        "console_scripts": ["micado=micado.cli:cli"],
    },
)
